#!python
#
# This file is part of aafigure. https://github.com/aafigure/aafigure
# (C) 2007 Chris Liechti <cliechti@gmx.net>
#
# SPDX-License-Identifier:    BSD-3-Clause
"""
AAFigure package.
"""

from .aafigure import process, render, UnsupportedFormatError, AsciiArtImage
