#!python
#
# This file is part of aafigure. https://github.com/aafigure/aafigure
# (C) 2009 Chris Liechti <cliechti@gmx.net>
#
# SPDX-License-Identifier:    BSD-3-Clause
"""\
Exception classes for the aafigure package.
"""

class UnsupportedFormatError(Exception):
    """Error message when an unknown format is specified"""
